import functools
from .aes import aes_decrypt, aes_encrypt
from .ecies import encryptedECIES, decryptedECIES, privateKey2publicKey
from . import cache
from .errors import ParameterError, NotKeyError
import copy
import logging


class SecurityService:
    def __init__(self, private_key=None, public_key=None):
        self.private_key = private_key
        self.public_key = public_key

    def ecies_encrypt(self, plain_text: str) -> str:
        if not self.public_key:
            logging.warning('public_key is None')
            return plain_text
        if not plain_text or len(plain_text) < 4 or '*' in plain_text:
            return plain_text
        if ('X' in plain_text and plain_text[-1] != 'X') or ('x' in plain_text and plain_text[-1] != 'x'):
            return plain_text
        cipher_text = encryptedECIES(plain_text, self.public_key)
        if cipher_text.startswith('0x'):
            cipher_text = cipher_text[2:]
        return f"JM_{cipher_text}_{plain_text[-4:]}"

    def ecies_decrypt(self, cipher_text: str) -> str:
        if not self.private_key:
            logging.warning('private_key is None')
            return cipher_text
        if not cipher_text.startswith('JM_'):
            return cipher_text
        _cipher_text = cipher_text.split('_')[1]
        return decryptedECIES(_cipher_text, self.private_key)


__keys_cache = None


def init_cache(host, cache_type='LocalCache', cache_prefix='doubao-keys-', cache_instance=None, cache_uri=None, pre_cache=False, clean=True):
    global __keys_cache
    if not __keys_cache:
        cls = getattr(cache, cache_type)
        __keys_cache = cls(host, cache_instance=cache_instance, cache_prefix=cache_prefix, cache_uri=cache_uri, pre_cache=pre_cache, clean=clean)


def doubao_ecies_encrypt(data, fields=None, cid=None, gid=None, agent_id=None):
    global __keys_cache

    def _encrypt(text, public_key):
        security_service = SecurityService(public_key=public_key)
        return security_service.ecies_encrypt(text)

    def _encrypt_fields(d, public_key):
        _data = copy.deepcopy(d)
        for field in fields:
            field_list = field.split('.')
            if len(field_list) == 1:
                if field_list[0] in _data:
                    _data[field_list[0]] = _encrypt(_data[field_list[0]], public_key)
            elif len(field_list) > 1:
                try:
                    code_str = '_data' + ''.join(["['{}']".format(x) if isinstance(x, str) else '[{}]'.format(x) for x in field_list])
                    v = eval(code_str)
                    code_str = '_data' + ''.join(["['{}']".format(x) if isinstance(x, str) else '[{}]'.format(x) for x in field_list[:-1]])
                    dv = {field_list[-1]: _encrypt(v, public_key)}
                    eval("{code_str}.update({dv})".format(code_str=code_str, dv=dv))
                except Exception as e:
                    logging.warning('field[{}] not exists'.format(field))
        return _data

    if not fields and not isinstance(data, str):
        return data
    if not cid:
        raise ParameterError('need kwargs `cid`')
    keys = __keys_cache.get_keys(cid, gid, agent_id)
    if not keys:
        raise NotKeyError('can not get keys form cid({cid}), gid({gid}), agent_id({agent_id})'.format(cid=cid, gid=gid, agent_id=agent_id))

    if isinstance(data, dict):
        return _encrypt_fields(data, keys['publicKey'])
    elif isinstance(data, list):
        return [_encrypt_fields(item, keys['publicKey']) for item in data]
    elif isinstance(data, str):
        return _encrypt(data, keys['publicKey'])
    else:
        return data


def safe_doubao_ecies_encrypt(data, fields=None, cid=None, gid=None, agent_id=None):
    try:
        _data = doubao_ecies_encrypt(data, fields=fields, cid=cid, gid=gid, agent_id=agent_id)
    except Exception as e:
        return False, str(e), data
    return True, "", _data


def doubao_privacy_policy(fields=None, cid=None, gid=None, agent_id=None):
    def wrapper(func):
        @functools.wraps(func)
        def _wrapper(data, *args, **kwargs):
            _data = doubao_ecies_encrypt(data, fields=fields, cid=cid, gid=gid, agent_id=agent_id)
            return func(_data, *args, **kwargs)
        return _wrapper
    return wrapper



from Crypto.Cipher import AES
import base64


def add_16(par):
    par = par.encode('utf-8')
    while len(par) % 16 != 0:
        par += b'\x00'
    return par


def aes_encrypt(text, key):
    text = add_16(text)
    aes = AES.new(add_16(key), AES.MODE_ECB)
    encrypt_text = aes.encrypt(text)
    return base64.encodebytes(encrypt_text).decode().strip()


def aes_decrypt(text, key):
    text = base64.decodebytes(text.encode('utf-8'))
    aes = AES.new(add_16(key), AES.MODE_ECB)
    decrypt_text = aes.decrypt(text)
    return decrypt_text.decode('utf-8').strip('\0')



from ecies.utils import generate_eth_key, generate_key, hex2prv
from ecies import encrypt, decrypt
from eth_keys import keys
from coincurve.utils import get_valid_secret
from eth_utils import decode_hex, encode_hex


def privateKey2publicKey(private_key):
    # 将私钥解析为操作格式
    priv_handler = keys.PrivateKey(decode_hex(private_key))
    # 获取公钥
    pub_handler = priv_handler.public_key
    pub_key = pub_handler.to_compressed_bytes()
    # 未压缩的公钥
    # pub_key.to_hex()
    return encode_hex(pub_key)


def encryptedECIES(msg, pub_key):
    return encode_hex(
        encrypt(pub_key, msg.encode('utf-8'))
    )


def decryptedECIES(enmsg, priv_key):
    return decrypt(
        priv_key, decode_hex(enmsg)
    ).decode('utf-8')



import abc
import requests
import redis
import json
from .errors import Api2Error
import logging

__all__ = ['LocalCache', 'RedisCache']


class BaseCache(metaclass=abc.ABCMeta):
    def __init__(self, api2_host, **kwargs):
        self.api2_host = api2_host
        pre_cache = kwargs.pop('pre_cache', None)
        if pre_cache:
            self.pre_cache()

    @abc.abstractmethod
    def get(self, key):
        pass

    @abc.abstractmethod
    def set(self, key, value):
        pass

    @abc.abstractmethod
    def clean(self):
        pass

    def get_keys(self, cid, gid, agent_id):
        p = [cid, gid, agent_id]
        key = '_'.join([str(x) for x in p if x is not None])
        _keys = self.get(key)
        if not _keys:
            _keys = self.get_data_from_remote(key)
        return _keys

    def pre_cache(self):
        rep = requests.get('{}/security-service-api/digicloud/encDecAnn/v1/encDec'.format(self.api2_host))
        result = rep.json()
        if result['errcode'] != 200:
            raise Api2Error(result['errmsg'])
        for item in result['data']:
            self.set(item['propertiesName'], item['propertiesValue']['keys'])

    def get_data_from_remote(self, key):
        rep = requests.get('{}/security-service-api/digicloud/encDecAnn/v1/encDec'.format(self.api2_host), params={'key': key})
        result = rep.json()
        if result['errcode'] != 200:
            raise Api2Error(result['errmsg'])
        if len(result['data']) == 0:
            logging.warning('can not find key {key}'.format(key=key))
            return None
        result = result['data'][0]
        self.set(result['propertiesName'], result['propertiesValue']['keys'])
        return result['propertiesValue']['keys']


class LocalCache(BaseCache):

    def __init__(self, api2_host, **kwargs):
        super().__init__(api2_host, **kwargs)
        self._cache = {}

    def get(self, key):
        return self._cache.get(key, None)

    def set(self, key, value):
        self._cache[key] = value

    def clean(self):
        self._cache = {}


class RedisCache(BaseCache):

    def __init__(self, api2_host, **kwargs):
        instance = kwargs.pop('cache_instance', None)
        if not instance:
            url = kwargs.pop('cache_uri', None) or 'redis://localhost/0'
            instance = redis.Redis.from_url(url)
        self.instance = instance
        self.cache_prefix = kwargs.pop('cache_prefix', 'doubao-keys-')
        is_clean = kwargs.pop('clean', False)
        if is_clean:
            self.clean()
        super().__init__(api2_host, **kwargs)

    def get_key(self, key):
        return '{}{}'.format(self.cache_prefix, key)

    def get(self, key):
        value = self.instance.get(self.get_key(key))
        return json.loads(value) if value else value

    def set(self, key, value):
        self.instance.set(self.get_key(key), json.dumps(value))

    def clean(self):
        for x in self.instance.scan_iter(match='{}*'.format(self.cache_prefix)):
            self.instance.delete(x)



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()


setup(
    name='doubao_crypt',
    version='1.1.1',
    author='biao.xu',
    author_email='biao.xu@baodanyun-inc.com',
    description='doubao crypt python library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.5.3",
    install_requires=['eciespy', 'requests', 'redis'],
    packages=['doubao_crypt'],
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)
